%PlotRunHistory: Generate plots illustrating the  performance
%%
%% DSA-BD:
%% C. Ortiz
%% Last modified: 3/31/2011
%%********************************************************************
function [h,legends] = DSA_BD_PlotRunHistory(h,legends,runhist, OPTIONS)
dimtext = [' (n=',num2str(OPTIONS.n),',m=',num2str(OPTIONS.m),')'];
legend_append =  DSA_BD_LegendAppend( OPTIONS);
objhist = [];
if isfield(runhist,'pobj')
    objhist(1,:) = runhist.pobj;
    objhist(2,:) = runhist.dobj;
end

resnorm = [];
if isfield(runhist,'resnorm')
    resnorm = runhist.resnorm;
end

cputime = [];
if isfield(runhist,'cputime')
    cputime = runhist.cputime;
end

feashist = [];
if isfield(runhist,'pfeas') && isfield(runhist,'dfeas')
    feashist(1,:) = runhist.pfeas;
    feashist(2,:) = runhist.dfeas;
end


epshist = [];
if isfield(runhist,'epsilon')
    epshist = runhist.epsilon;
end

lambdaratio = [];
if isfield(runhist,'lambdaratio')
    lambdaratio = runhist.lambdaratio;
end

lambdahist = [];
if isfield(runhist,'plambda')
    lambdahist = sqrt(runhist.dlambda./runhist.plambda);
end
[h,legends] = plotvalues(h,legends,legend_append,objhist,resnorm,feashist, ...
    epshist, lambdaratio,lambdahist,dimtext,cputime);
end


function  [h,legends] = plotvalues(h,legends,legend_append,Obj,Res,Feas,Eps,lRatio,lambdas,dtext,ctime)

%clf;
set(0,'defaultaxesfontsize',12);
set(0,'defaultlinemarkersize',2);
set(0,'defaulttextfontsize',12);
group = 0;
lines{1} = '-g';
lines{2} = '-k';
lines{3} = '--r';
lines{4} = '--b';
lines{5} = '-.y';
lines{6} = '-.c';

[h,legends] = plotPrimalDual(nargin,1,Obj,'duality gap',group,h,legends,legend_append,lines,dtext,'# iterations','{ p_{OBJ} & d_{OBJ} }');
%%
%% get axis scale for plotting residual
%%
[h,legends] = plotNorm(nargin,2,Res,'norm of residual measure',group,h,'point',legends,legend_append,lines,dtext,'# iterations','(p_{err}^2+d_{err}^2)^{1/2}');

%%
%% get axis scale for plotting infeasibiliy
%%
[h,legends] = plotPrimalDual(nargin,3,Feas,'infeasibility measure',group,h,legends,legend_append,lines,dtext);

%%
%% get axis scale for plotting epsilon of aprox subdifferential
%%
[h,legends] = plotNorm(nargin,4,Eps,'\epsilon-subdifferential param',group,h,'point',legends,legend_append,lines,dtext);

%%
%% get axis scale for plotting lambda ratio for adaptive lambda
%%
[h,legends] = plotNormNoLog(nargin,5,lRatio,'Adaptive lambda ratio: \lambda_k/\lambda_0',group,h,'ratio',legends,legend_append,lines,dtext);

%%
%% get axis scale for plotting wieghted lambdas
%%
[h,legends] = plotNormNoLog(nargin,6,lambdas,'weigthed lambdas (\theta)',group,h,'\theta',legends,legend_append,lines,dtext);
%%
%% get axis scale for plotting residual vs time
%%
[h,legends] = plotNormVsTime(nargin,7,Res,'norm of residual measure in time',group,h,'point',legends,legend_append,lines,dtext,ctime,'time (secs)','(p_{err}^2+d_{err}^2)^{1/2}');

%%
%% get axis scale for plotting -log10(|p_obj-dobj|) vs time
%%
[h,legends] = plotAbsGapVsTime(nargin,8,Obj,'norm of gap in time',group,h,'point',legends,legend_append,lines,dtext,ctime,'time (secs)','-log10(|p_{OBJ} - d_{OBJ}|)');
end

function [h,legends] = plotPrimalDual(narg,arg,arrays,titl,group,h,legends,legend_append,lines,dtext,xtext,ytext)
if narg >= arg && ~isempty(arrays);
    figure(arg+group);
    if isempty(legends) || length(legends)<arg || isempty(legends{arg})
        lgnd{1}=['primal ',legend_append];
        lgnd{2}=['dual ',legend_append];
        legends{arg}=lgnd;
        index = 1;
        hold off;
    else
        lgnd = legends{arg};
        index = length(lgnd)+1;
        lgnd{index}=['primal ',legend_append];
        lgnd{index+1}=['dual ',legend_append];
        legends{arg}=lgnd;
        hold on;
    end;

    index2 = length(h)+1;
    h(index2)= plot(1:size(arrays,2),arrays(1,:),lines{index});
    hold on;
    h(index2+1) = plot(1:size(arrays,2),arrays(2,:),lines{index+1});
    title([titl,dtext]);
    if nargin>=11; SetAxis(xtext,ytext); end;
    SetLegend(lgnd);
    hold off;
end
end

function [h,legends] = plotNorm(narg,arg,arrays,titl,group,h,newlegend,legends,legend_append,lines,dtext,xtext,ytext)
if narg >= arg && ~isempty(arrays);
    figure(arg+group);
    if isempty(legends) || length(legends)<arg || isempty(legends{arg})
        lgnd{1}=[newlegend,' ',legend_append];
        legends{arg}=lgnd;
        index = 1;
        hold off;
    else
        lgnd = legends{arg};
        index = length(lgnd)+1;
        lgnd{index}=[newlegend,' ',legend_append];
        legends{arg}=lgnd;
        hold on;
    end;
    index2 = length(h)+1;
    h(index2) = plot(1:size(arrays,2),-log10(arrays),lines{index});
    title(['-LOG10  ',titl,dtext]);
    if nargin>=13; SetAxis(xtext,ytext); end;
    SetLegend(lgnd);
    hold off;
end
end

function [h,legends] = plotNormVsTime(narg,arg,arrays,titl,group,h,newlegend,legends,legend_append,lines,dtext,ctime,xtext,ytext)
if narg >= arg && ~isempty(arrays)&& ~isempty(ctime);
    figure(arg+group);
    legend_append = [legend_append,' (it=',num2str(length(ctime)),')'];
    if isempty(legends) || length(legends)<arg || isempty(legends{arg})
        lgnd{1}=[newlegend,' ',legend_append];
        legends{arg}=lgnd;
        index = 1;
        hold off;
    else
        lgnd = legends{arg};
        index = length(lgnd)+1;
        lgnd{index}=[newlegend,' ',legend_append];
        legends{arg}=lgnd;
        hold on;
    end;
    index2 = length(h)+1;
    h(index2) = plot([0,ctime],[0,-log10(arrays)],lines{index});
    title(['-LOG10  ',titl,dtext]);
    if nargin>=14; SetAxis(xtext,ytext); end;
    SetLegend(lgnd);
    hold off;
end
end
function [h,legends] = plotAbsGapVsTime(narg,arg,arrays,titl,group,h,newlegend,legends,legend_append,lines,dtext,ctime,xtext,ytext)
if narg >= arg && ~isempty(arrays)&& ~isempty(ctime);
    figure(arg+group);
    legend_append = [legend_append,' (it=',num2str(length(ctime)),')'];
    if isempty(legends) || length(legends)<arg || isempty(legends{arg})
        lgnd{1}=[newlegend,' ',legend_append];
        legends{arg}=lgnd;
        index = 1;
        hold off;
    else
        lgnd = legends{arg};
        index = length(lgnd)+1;
        lgnd{index}=[newlegend,' ',legend_append];
        legends{arg}=lgnd;
        hold on;
    end;
    index2 = length(h)+1;
    h(index2) = plot([0,ctime],[0,-log10(abs(arrays(1,:)-arrays(2,:))./(1+abs(arrays(1,:))+abs(arrays(2,:))))],lines{index});
    title(['-LOG10  ',titl,dtext]);
    if nargin>=14; SetAxis(xtext,ytext); end;
    SetLegend(lgnd);
    hold off;
end
end

function [h,legends] = plotNormNoLog(narg,arg,arrays,titl,group,h,newlegend,legends,legend_append,lines,dtext)
if narg >= arg && ~isempty(arrays);
    figure(arg+group);
    if isempty(legends) || length(legends)<arg || isempty(legends{arg})
        lgnd{1}=[newlegend,' ',legend_append];
        legends{arg}=lgnd;
        index = 1;
        hold off;
    else
        lgnd = legends{arg};
        index = length(lgnd)+1;
        lgnd{index}=[newlegend,' ',legend_append];
        legends{arg}=lgnd;
        hold on;
    end;
    index2 = length(h)+1;
    h(index2) = plot(1:size(arrays,2),arrays,lines{index});
    title([titl,dtext]);
    SetLegend(lgnd);
    hold off;
end
end

function SetLegend(lgnd)
s = ['''',lgnd{1},''''];
for i = 2 : length(lgnd)
    s = [s,',''',lgnd{i},''''];
end
s = ['legend(',s,')'];
eval(s);
end

function SetAxis(xtext,ytext)
xlabel(xtext);
ylabel(ytext);
end
