% DSA_BD_normOfOperator: compute the first singular value of At
%%
%% DSA_BD:
%% C. Ortiz
%% Last Modified 3/31/2011
%% Last Modified 11/27/2011
%%***********************************************************

function [L] = DSA_BD_normOfOperator(blk,At,tol,m)


if nargin < 4
    tol = 1e-6;
end
if nargin < 5
    m = size(At{1},2);
end

fprintf('\n Calculating Norm of A');
MAXITER = 2000;
TOLERANCE = min(1e-6,tol);
try
    AAt = sparse(m,m);
    for p = 1:size(blk,1)
        AAt = AAt + At{p,1}'*At{p,1};
    end
    pertdiag = 1e-13*ones(m,1);
    AAt = AAt + spdiags(pertdiag,0,m,m);
    options.disp = 0;
    options.maxit = MAXITER;
    options.tol = TOLERANCE;
    L = sqrt(svds(AAt,1,'L',options));

    fprintf('\nNorm of A = %d\n',L);
    return;
catch e
    fprintf('\nsvds(A,1,L,options) is not used.' );
end

q = rand(size(At{1},2),1);%q = rand(size(A,1),1);
q = q/norm(q);
Atq = DSA_BD_Atyfun(blk,At,q);%Atyfun3(At,q,n);
L1 = 0;
L2 = 1;
q1 = q;

for i = 1:MAXITER
    z =DSA_BD_AXfun(blk,At,Atq);% AXfun(blk,At,[],Atq);%AXfun3(A,Atq);
    q = z/norm(z);
    Atq = DSA_BD_Atyfun(blk,At,q);%Atyfun(blk,At,[],[],q);%Atyfun3(At,q,n);
    L1 = q'*DSA_BD_AXfun(blk,At,Atq);%AXfun(blk,At,[],Atq);%AXfun3(A,Atq);
    error1 = abs(L1 - L2);
    error2 = norm(q - q1);
    if i > 1 && error1 < TOLERANCE && error2 < TOLERANCE
        break;
    end
    q1 = q;
    L2 = L1;
end
L =sqrt(L1);
if (i == MAXITER)
    fprintf('WARNING: Power iteration did not converged. Errors: %f %f',error1,error2);
end
fprintf('\nPower iterations = %d',i);
fprintf('\nNorm of A = %d\n',L);
end
