%Basic Operations: Perform various basic operations for block structured elements
%%******************************************************************
%% ops:
%%
%%   Z = DSA_BD_ops(X,operand,Y,alpha);
%%
%%  INPUT:        X = a matrix or a scalar
%%                    or a CELL ARRAY consisting only of matrices
%%          operand = sym, transpose, triu, tril,
%%                    real, imag, sqrt, abs, max, min, nnz,
%%                    spdiags, ones, norm, sum, row-norm,
%%                    rank1, rank1inv, inv
%%                    +,  -, *, .*,  ./, .^
%%     Y (optional) = a matrix or a scalar
%%                    or a CELL ARRAY consisting only of matrices
%%  alpha(optional) = a matrix or a scalar
%%                    or a CELL ARRAY consisting only of matrices
%%    X2 (optional) = a matrix or a scalar
%%                    or a CELL ARRAY consisting only of matrices
%%    Y2 (optional) = a matrix or a scalar
%%                    or a CELL ARRAY consisting only of matrices
%% alpha2(optional) = a matrix or a scalar
%%                    or a CELL ARRAY consisting only of matrices
%%***********************************************************
%% DSA_BD:
%% Modified by C.Ortiz
%% Last Modified: 3/31/2011

function Z = DSA_BD_ops(X,operand,Y,alpha,X2,Y2,alpha2)


if (nargin == 2)
    if strcmp(operand,'norm');
        if ~iscell(X);
            Z = full(norm(X,'fro'));%full(sqrt(sum(sum(X.*X))));
        else
            Z = 0;
            for p = 1:length(X); Z = Z + norm(X{p},'fro')^2;end;%sum(sum(X{p}.*X{p})); end;
            Z = sqrt(Z);
        end
        
    elseif strcmp(operand,'zeros')
        if ~iscell(X)
            Z = sparse(size(X,1),size(X,2));
        else
            Z = cell(size(X));
            for p = 1:length(X)
                Z{p} = sparse(size(X{p},1),size(X{p},2));
            end
        end
    end
end
%%
if (nargin == 3)
    if strcmp(operand,'+') || strcmp(operand,'-') || ...
            strcmp(operand,'/') || strcmp(operand,'./') || ...
            strcmp(operand,'*') || strcmp(operand,'.*') || ...
            strcmp(operand,'.^')
        if (~iscell(X) && ~iscell(Y))
            eval(['Z = X',operand,'Y;']);
        elseif (iscell(X) && iscell(Y))
            Z = cell(size(X));
            for p = 1:length(X)
                if (size(X{p},2) == 1) && (size(Y{p},2) == 1) && ...
                        (strcmp(operand,'*') || strcmp(operand,'/'))
                    eval(['Z{p} = X{p}.',operand,'Y{p};']);
                else
                    eval(['Z{p} = X{p} ',operand,'Y{p};']);
                end;
            end;
        elseif (iscell(X) && ~iscell(Y))
            Z = cell(size(X));
            for p = 1:length(X);
                eval(['Z{p} = X{p}',operand,'Y;']);
            end
        elseif (~iscell(X) && iscell(Y))
            Z = cell(size(Y));
            for p = 1:length(Y);
                eval(['Z{p} = X',operand,'Y{p};']);
            end
        end
    elseif strcmp(operand,'diffnorm')
        if ~iscell(X);
            Z = full(norm(X-Y,'fro'));%full(sqrt(sum(sum(X.*X))));
        else
            Z = 0;
            for p = 1:length(X); Z = Z + norm(X{p}-Y{p},'fro')^2;end;%sum(sum(X{p}.*X{p})); end;
            Z = sqrt(Z);
        end
    else
        error([operand,' is not available, check input arguments']);
    end
end
%%
if (nargin == 4)
    if strcmp(operand,'calc2')
        if (~iscell(X) )
            Z = -X+Y+alpha;
        else 
            Z = cell(size(X));
            for p = 1:length(X)
                Z{p} = -X{p}+Y{p}+alpha{p};
            end;
        end
        
    elseif strcmp(operand,'calc8')
        if (~iscell(X) )
            Z = (X-Y)/alpha;
        else 
            Z = cell(size(X));
            for p = 1:length(X)
                Z{p} = (X{p}-Y{p})/alpha;
            end;
        end
    else   
        error([operand,' is not available']);
    end
end

if (nargin == 5)
    if strcmp(operand,'calc3')
       if (~iscell(X) )
            Z = X-((Y-X2)*alpha);
        else 
            Z = cell(size(X));
            for p = 1:length(X)
                Z{p} = X{p}-((Y{p}-X2{p})*alpha);
            end;
        end 
    
    elseif strcmp(operand,'calc5')
       if (~iscell(X) )
            Z = X-((Y-X2)*alpha);
        else 
            Z = cell(size(X));
            for p = 1:length(X)
                Z{p} = X{p}-((Y{p}-X2{p})*alpha);
            end;
        end 

    elseif strcmp(operand,'calc7')
       if (~iscell(X) )
            Z = (1-Y/alpha)*X+(Y/alpha)*X2;
        else 
            Z = cell(size(X));
            for p = 1:length(X)
                Z{p} = (1-Y/alpha)*X{p}+(Y/alpha)*X2{p};
            end;
        end 

    end
end

if (nargin == 6)
    if strcmp(operand,'calc1')
        if (~iscell(X) )
            Z = (((X-Y)/alpha) + X2-Y2);
        else 
            Z = cell(size(X));
            for p = 1:length(X)
                Z{p} = (((X{p}-Y{p})/alpha)+X2{p}-Y2{p});
            end;
        end
    elseif strcmp(operand,'calc4')
        if (~iscell(X) )
            Z = (X/Y)+(((X2-Y2)/alpha));
        else 
            Z = cell(size(X));
            for p = 1:length(X)
                Z{p} = (X{p}/Y)+(((X2{p}-Y2{p})/alpha));
            end;
        end
    end
end

if (nargin == 7)
    if strcmp(operand,'calc6')
        if (~iscell(X) )
            Z = ((X*(1-Y/alpha))+((X2*Y/alpha)-(Y2*alpha2)));
        else
            Z = cell(size(X));
            for p = 1:length(X)
                Z{p} = ((X{p}*(1-Y/alpha))+((X2{p}*Y/alpha)-(Y2{p}*alpha2)));
            end;
        end
    end
end
%%******************************************************************
