
#if !defined(MAX)
#define  MAX(A, B)   ((A) > (B) ? (A) : (B))
#endif

#if !defined(MIN)
#define  MIN(A, B)   ((A) < (B) ? (A) : (B))
#endif

#if !defined(SQR)
#define SQR(x) ((x)*(x))
#endif

/********************************
* dzeros: set x(0:n-1) to zero. 
*********************************/
void dzeros(double *x, const int n)

{ int i; 

  for (i=0; i<n-3; i++) {
     x[i] = 0; i++; 
     x[i] = 0; i++; 
     x[i] = 0; i++; 
     x[i] = 0; }
  if (i<n-1) { 
     x[i] = 0; i++; 
     x[i] = 0; i++; }
  if (i<n) {
     x[i] = 0; }
  return; 
}
/********************************
  dcopy(x,y,n) 
  copy x(0:n-1) to y(0:n-1) 
*********************************/
void dcopy(const double *x, double *y, const int n)

{ int i; 

  for (i=0; i<n-3; i++) {
     y[i] = x[i]; i++; 
     y[i] = x[i]; i++; 
     y[i] = x[i]; i++; 
     y[i] = x[i]; }
  if (i<n-1) { 
     y[i] = x[i]; i++; 
     y[i] = x[i]; i++; }
  if (i<n) {
     y[i] = x[i]; }
  return; 
}
/********************************
* realdot: x dense vector,  
*          y dense vector
*********************************/
double realdot(const double *x, const double *y, const int n)

{  int i;
   double r; 

   r=0.0;
   for (i=0; i<n-3; i++) {             /* LEVEL 4 */
      r += x[i] * y[i]; i++;
      r += x[i] * y[i]; i++;
      r += x[i] * y[i]; i++;
      r += x[i] * y[i]; }
   if (i<n-1) {                        /* LEVEL 2 */
      r += x[i] * y[i]; i++;
      r += x[i] * y[i]; i++; }
   if (i<n) {                          /* LEVEL 1 */
      r += x[i] * y[i]; }
   return r; 
}
/********************************
* realdotde: x dense matrix,  
*            y dense vector
*********************************/
double realdotde(const double *x, const int col, 
                 const double *y, const int n)

{  int i, idx;
   double r; 

   idx=col*n; 
   r=0.0;
   for (i=0; i<n-3; i++) {             /* LEVEL 4 */
      r += x[i+idx] * y[i]; i++;
      r += x[i+idx] * y[i]; i++;
      r += x[i+idx] * y[i]; i++;
      r += x[i+idx] * y[i]; }
   if (i<n-1) {                        /* LEVEL 2 */
      r += x[i+idx] * y[i]; i++;
      r += x[i+idx] * y[i]; i++; }
   if (i<n) {                          /* LEVEL 1 */
      r += x[i+idx] * y[i]; }
   return r; 
}
/********************************
* realdotsp: x sparse matrix, 
*            y dense  vector
*********************************/
double realdotsp(const double *x, const mwIndex *irx, const mwIndex *jcx, 
                const int col, const double *y)

{  int i, row, idxstart, idxend;
   double r; 

   idxstart=jcx[col]; idxend=jcx[col+1]; 
   r=0.0;
   for (i=idxstart; i<idxend-3; i++) {      /* LEVEL 4 */
      row = irx[i]; r += x[i]*y[row]; i++;
      row = irx[i]; r += x[i]*y[row]; i++;
      row = irx[i]; r += x[i]*y[row]; i++;
      row = irx[i]; r += x[i]*y[row]; }
   if (i<idxend-1) {                        /* LEVEL 2 */
      row = irx[i]; r += x[i]*y[row]; i++;
      row = irx[i]; r += x[i]*y[row]; i++; }
   if (i<idxend) {                          /* LEVEL 1 */
      row = irx[i]; r += x[i]*y[row]; }
   return r; 
}
/********************************
* realdotmat: x dense matrix, 
*             y dense matrix   
*********************************/
double realdotmat(const double *x, const int idx1, const double *y, 
                  const int idx2,  const int n)

{  int i;
   double r; 

   r=0.0;
   for(i=0; i< n-3; i++){                 /* LEVEL 4 */
      r+= x[i+idx1] * y[i+idx2]; i++;
      r+= x[i+idx1] * y[i+idx2]; i++;
      r+= x[i+idx1] * y[i+idx2]; i++;
      r+= x[i+idx1] * y[i+idx2]; 
   }
   if(i < n-1){                           /* LEVEL 2 */
      r+= x[i+idx1] * y[i+idx2]; i++;
      r+= x[i+idx1] * y[i+idx2]; i++;
   }
   if(i < n){                             /* LEVEL 1 */
      r+= x[i+idx1] * y[i+idx2]; 
   }
   return r; 
}
/********************************
* saxpymat:  z = alpha*y + z
* z dense matrix, y dense matrix
********************************/
void saxpymat(const double alp, const double *y, const int idx1, 
              double *z, const int idx2, const int istart, const int iend)
{  int i;
  
   for(i=istart; i< iend-3; i++){             /* LEVEL 4 */
      z[i+idx2] += alp * y[i+idx1]; i++;
      z[i+idx2] += alp * y[i+idx1]; i++;
      z[i+idx2] += alp * y[i+idx1]; i++;
      z[i+idx2] += alp * y[i+idx1]; 
   }
   if(i < iend-1){                            /* LEVEL 2 */
      z[i+idx2] += alp * y[i+idx1]; i++;
      z[i+idx2] += alp * y[i+idx1]; i++;
   }
   if(i < iend){                              /* LEVEL 1 */
      z[i+idx2] += alp * y[i+idx1]; 
   }
   return;
}
/********************************
  saxpy(z, alp, y, n)
  Computes  z = z + alp*y 
  z dense vector, y dense vector
********************************/
void saxpy(double *z, const double alp, 
           const double *y, const int n)
{  int i;
  
   for(i=0; i< n-3; i++){                /* LEVEL 4 */
      z[i] += alp * y[i]; i++;
      z[i] += alp * y[i]; i++;
      z[i] += alp * y[i]; i++;
      z[i] += alp * y[i]; 
   }
   if(i < n-1){                          /* LEVEL 2 */
      z[i] += alp * y[i]; i++;
      z[i] += alp * y[i]; i++;
   }
   if(i < n){                            /* LEVEL 1 */
      z[i] += alp * y[i]; 
   }
   return;
}
/********************************
  sxpay(y, alp, z, n)
  Computes y = z + alp*y 
  z dense vector, y dense vector
********************************/
void sxpay(double *y, const double alp, 
           const double *z, const int n)
{  int i;
  
   for(i=0; i< n-3; i++){                 /* LEVEL 4 */
      y[i] = z[i] + alp*y[i]; i++;
      y[i] = z[i] + alp*y[i]; i++;
      y[i] = z[i] + alp*y[i]; i++;
      y[i] = z[i] + alp*y[i]; 
   }
   if(i < n-1){                           /* LEVEL 2 */
      y[i] = z[i] + alp*y[i]; i++;
      y[i] = z[i] + alp*y[i]; i++;
   }
   if(i < n){                             /* LEVEL 1 */
      y[i] = z[i] + alp*y[i]; 
   }
   return;
}
/********************************
* compute A*x
* x dense
* Ax must be initialized to be 0
*********************************/
void compAx(int mA, int nA, double *A, mwIndex *irA, mwIndex *jcA, 
            int isspA, double *x, double *Ax)
                     
{  int	  i, j, r, istart, iend;
   double tmp;
      
   dzeros(Ax,mA); 
   if (!isspA) {
      for (j=0; j<nA; j++){ saxpymat(x[j],A,j*mA,Ax,0,0,mA); } 
   } else {
      for (j=0; j<nA; j++){
          tmp = x[j];
          istart = jcA[j]; iend = jcA[j+1]; 
          if (tmp != 0) {
  	     for (i=istart; i<iend; i++) {
                 r = irA[i]; 
	         Ax[r] += tmp*A[i]; }
	  }
      }
   }
   return;
}
/**********************************************************
* smat: B dense
**********************************************************/
void smat(int n, const double *A, double *B)

{  int idx, i, j, jn, k;
   double ir2=1/sqrt(2.0); 

      idx = 0; 
      for (j=0; j<n; j++) { 
          jn = j*n; 
          for (i=0; i<j; i++) { 
              B[i+jn] = ir2*A[idx]; 
              idx++; } 
          B[j+jn] = A[idx];
          idx++; 
      }
      for (j=0; j<n; ++j){
         jn = j*n; 
         for (k=0; k<j ; ++k){ B[j+k*n] = B[k+jn]; }
      }
return; 
}
/**********************************************************
* svec: B dense
**********************************************************/
void svec(int n, double *A, double *B) 

{  int idx, i, j, jn, k, kstart, kend;
   double r2=sqrt(2.0); 

   idx = 0; 
   for (j=0; j<n; j++) { 
      jn = j*n; 
      for (i=0; i<j; i++) { B[idx] = A[i+jn]*r2; idx++; } 
      B[idx] = A[j+jn]; idx++; 
   }
return;
}
/**********************************************************
* svecadd: B dense
*          DO NOT intialize B to 0.
* B = B + alpha*svec(A); 
**********************************************************/
void svecadd(int n, double *A, double *B, double alpha) 

{  int idx, i, j, jn, k, kstart, kend;
   double r2; 

   r2 = alpha*sqrt(2.0); 
   idx = 0; 
   for (j=0; j<n; j++) { 
      jn = j*n; 
      for (i=0; i<j; i++) { B[idx] += A[i+jn]*r2; idx++; } 
      B[idx] += alpha*A[j+jn]; idx++; 
   }
return;
}
/**********************************************************
* compute inner product
**********************************************************/
void inprod(int mA, int nA, double *A, mwIndex *irA, mwIndex *jcA, 
            int isspA, double *x, double *Ax)

{  int j; 

   if (isspA) { 
      for (j=0; j<nA; j++){ Ax[j] = realdotsp(A,irA,jcA,j,x); }
   } else {
      for (j=0; j<nA; j++){ Ax[j] = realdotde(A,j,x,mA); }	  
   }
return;
}
/**********************************************************/
