/***********************************************************************
* mexpsqmr: 
*          
* [x,resnrm,solve_ok] = mexpsqmr(blk,A,b,P1,P2,Dsch12,...
*                       diagmatM,pertdiag,regparam,tol,maxit,stagnate_check);  
*
************************************************************************/

#include <mex.h>
#include <math.h>
#include <matrix.h> 

#include "mexaux.c"

#if !defined(MX_API_VER) || ( MX_API_VER < 0x07030000 )
typedef int mwIndex;
typedef int mwSize;
#endif

/********************************
* MATVEC_lin: linear block. 
* matvec <--- matvec + Ax
*********************************/
void MATVEC_lin(int mA1, int nA1, double *A1, 
                mwIndex *irA1, mwIndex *jcA1, int isspA1, 
                double *Dsch12_lin, double *tmpvec_lin, double *matvectmp,
                double *xvec, double *matvec, double regparam)

{   int j; 

    compAx(mA1, nA1, A1, irA1, jcA1, isspA1, xvec, tmpvec_lin); 
    for (j=0; j<mA1; j++) { 
        tmpvec_lin[j] *= Dsch12_lin[j];
    } 
    inprod(mA1, nA1, A1, irA1, jcA1, isspA1, tmpvec_lin, matvectmp);
    saxpy(matvec,regparam,matvectmp,nA1); 
return; 
}

/********************************
* MATVEC1: SDP block
* matvec <--- matvec + regparam*Ax
*********************************/
void MATVEC1(int blksize, 
             int mA, int nA, double *A, mwIndex *irA, mwIndex *jcA, int isspA, 
             int nP1, int nP2,             
             double *Atymat, 
             double *P1, double *P2, double *P1t, double *P2t, double *Dsch12, 
             double *T0, double *T1, double *T2, double *Tmp, double *tmpvec, double *matvectmp,
             double *xvec, double *matvec, double regparam)
                              
{   int      j, k, jm;
    char    *chn="N";
    double   one=1.0, zero=0.0; 

       /*--------- Atymat = smat(A*xvec) -------------*/
       compAx(mA, nA, A, irA, jcA, isspA, xvec, tmpvec);    
       smat(blksize,tmpvec,Atymat);         

       /*--------- T0=P1t*Atymat (nP1xblksize) -----*/
       dgemm_(chn,chn,&nP1,&blksize,&blksize,&one, P1t,&nP1, Atymat,&blksize, &zero, T0,&nP1); 

       /*--------- T1=(T0*P1)*P1t (nP1xblksize) -----*/
       dgemm_(chn,chn,&nP1,&nP1,&blksize,&one, T0,&nP1, P1,&blksize, &zero, Tmp,&nP1); 
       dgemm_(chn,chn,&nP1,&blksize,&nP1,&one, Tmp,&nP1, P1t,&nP1, &zero, T1,&nP1); 

       /*--------- Tmp=T0*P2 (nP1xnP2) -----*/
       dgemm_(chn,chn,&nP1,&nP2,&blksize,&one, T0,&nP1, P2,&blksize, &zero, Tmp,&nP1);

       for (j=0; j<nP2; j++) { 
	  jm = j*nP1; 
          for (k=0; k<nP1; k++) { Tmp[k+jm] *= Dsch12[k+jm]; }
       }        
       /*--------- T2=Tmp*P2t (nP1xblksize) -----*/
       dgemm_(chn,chn,&nP1,&blksize,&nP2,&one, Tmp,&nP1, P2t,&nP2, &zero, T2,&nP1); 

       for (j=0; j<blksize; j++) { 
	  jm = j*nP1; 
          for (k=0; k<nP1; k++) { T2[k+jm] += 0.5*T1[k+jm]; }
       }        
       /*--------- Atymat=P1*T2 (blksizexblksize) -----*/
       dgemm_(chn,chn,&blksize,&blksize,&nP1,&one, P1,&blksize, T2,&nP1, &zero, Atymat,&blksize);
               
       for (j=0; j<blksize; j++) { 
	  jm = j*blksize; 
          for (k=0; k<=j; k++) { Atymat[k+jm] += Atymat[j+k*blksize]; }
       }               
       /*--------- A'*svec(Atymat) -------------*/
       svec(blksize,Atymat,tmpvec);  
       inprod(mA, nA, A, irA, jcA, isspA, tmpvec, matvectmp);
       saxpy(matvec,regparam,matvectmp,nA); 
return;
}
/********************************
* MATVEC2: SDP block
* matvec <--- matvec + regparam*Ax
*********************************/
void MATVEC2(int blksize, 
             int mA, int nA, double *A, mwIndex *irA, mwIndex *jcA, int isspA, 
             int nP1, int nP2,             
             double *Atymat, 
             double *P1, double *P2, double *P1t, double *P2t, double *Dsch12, 
             double *T0, double *T1, double *T2, double *Tmp, double *tmpvec, double *matvectmp,
             double *xvec, double *matvec, double regparam)

                              
{   int      j, k, jm;
    char    *chn="N";
    double   one=1.0, zero=0.0; 

       /*--------- Atymat = smat(A*xvec) -------------*/
       compAx(mA, nA, A, irA, jcA, isspA, xvec, tmpvec);    
       smat(blksize,tmpvec,Atymat);         

       /*--------- T0=P2t*Atymat (nP2xblksize) -----*/
       dgemm_(chn,chn,&nP2,&blksize,&blksize,&one, P2t,&nP2, Atymat,&blksize, &zero, T0,&nP2); 

       /*--------- T1=(T0*P2)*P2t (nP2xblksize) -----*/
       dgemm_(chn,chn,&nP2,&nP2,&blksize,&one, T0,&nP2, P2,&blksize, &zero, Tmp,&nP2); 
       dgemm_(chn,chn,&nP2,&blksize,&nP2,&one, Tmp,&nP2, P2t,&nP2, &zero, T1,&nP2); 

       /*--------- Tmp=T0*P1 (nP2xnP1) -----*/
       dgemm_(chn,chn,&nP2,&nP1,&blksize,&one, T0,&nP2, P1,&blksize, &zero, Tmp,&nP2);

       for (j=0; j<nP1; j++) { 
	  jm = j*nP2; 
          for (k=0; k<nP2; k++) { Tmp[k+jm] *= (1-Dsch12[j+k*nP1]); }
       }        
       /*--------- T2=Tmp*P1t (nP2xblksize) -----*/
       dgemm_(chn,chn,&nP2,&blksize,&nP1,&one, Tmp,&nP2, P1t,&nP1, &zero, T2,&nP2); 

       for (j=0; j<blksize; j++) { 
	  jm = j*nP2; 
          for (k=0; k<nP2; k++) { T2[k+jm] += 0.5*T1[k+jm]; }
       }        
       /*--------- Atymat=P2*T2 (blksizexblksize) -----*/
       dgemm_(chn,chn,&blksize,&blksize,&nP2,&one, P2,&blksize, T2,&nP2, &zero, Atymat,&blksize);
               
       for (j=0; j<blksize; j++) { 
	  jm = j*blksize; 
          for (k=0; k<=j; k++) { Atymat[k+jm] += Atymat[j+k*blksize]; }
       }               
       /*--------- A'*svec(Atymat) -------------*/
       svecadd(blksize,Atymat,tmpvec,-1.0);  
       inprod(mA, nA, A, irA, jcA, isspA, tmpvec, matvectmp);
       saxpy(matvec,regparam,matvectmp,nA); 
return;
}
/********************************
* MATVEC3: SDP block
* matvec <--- matvec + regparam*Ax
*********************************/
void MATVEC3(int blksize, 
             int mA, int nA, double *A, mwIndex *irA, mwIndex *jcA, int isspA, 
             double *Atymat, 
             double *tmpvec, double *matvectmp,
             double *xvec, double *matvec, double regparam)

                              
{   int      j, k, jm;
    char    *chn="N";
    double   one=1.0, zero=0.0; 

       /*--------- Atymat = smat(A*xvec) -------------*/
       compAx(mA, nA, A, irA, jcA, isspA, xvec, tmpvec);    
       smat(blksize,tmpvec,Atymat);         

       /*--------- A'*svec(Atymat) -------------*/
       svec(blksize,Atymat,tmpvec);  
       inprod(mA, nA, A, irA, jcA, isspA, tmpvec, matvectmp);
       saxpy(matvec,regparam,matvectmp,nA); 
return;
}
/**********************************************************
* main
**********************************************************/
void mexFunction(
      int nlhs,   mxArray  *plhs[], 
      int nrhs,   const mxArray  *prhs[] )
    
{    const mxArray *L_FIELD;
     mxArray       *tmparr[7];  
     mxArray       *A_cell_pr, *blk_cell_pr;
     mxArray       *P1_cell_pr, *P2_cell_pr;
     mxArray       *Dsch12_cell_pr;  
     double        *b, *A, *P1, *P2, *P1t, *P2t, *solve_ok;
     double        *Dsch12, *diagmatM, *pertdiag; 
     double        *Atymat, *T0, *T1, *T2, *Tmp; 
     double        *tmpvec, *tmpvec_lin, *matvectmp; 
     double        *xvec, *rvec, *pvec, *zvec, *dvec, *res, *Ap, *Ad, *resnrm; 
     mwIndex       *irA, *jcA, *irA1, *jcA1;
     mwSize        mb, nb, mA, nA, isspA, index, numcell, existlinblk; 
     mwSize        mA1, nA1, isspA1; 

     double        *A1, *Dsch12_lin;    

     mwIndex   subs[2];
     mwSize    nsubs=2; 
     double    tol, regparam, TINY, one=1.0, zero=0.0; 
     double    tau_old, rho_old, theta_old, tau, rho, theta, sigma; 
     double    alpha, beta, gam, eta, err, ratio, maxratio, minratio; 
     int       nP1, nP2, nDsch12, nDsch12_lin, j, k, jm, iter, stagnate_check; 
     int       p, maxit, numblk, blksize, type; 

/***** get pointers *****/    

   if (nrhs <= 11) {
       mexErrMsgTxt("mexpsqmr: must have at least 11 inputs"); }
   if (nlhs > 3) { 
       mexErrMsgTxt("mexpsqmr: requires at most 3 output argument"); }
   numcell = mxGetM(prhs[1]);
   if (!mxIsCell(prhs[1])) {
      mexErrMsgTxt("mexpsqmr: At must be cell array"); }
   if (numcell == 2) {
      existlinblk = 1; 
   } else {
      existlinblk = 0; 
   }
   mb = mxGetM(prhs[2]); 
   nb = mxGetN(prhs[2]);            
   b  = mxGetPr(prhs[2]); 
   if (nb > 1) { 
      mexErrMsgTxt("mexpsqmr: rhs must be a column vector"); }
   if (mxIsSparse(prhs[2])) { 
      mexErrMsgTxt("mexpsqmr: sparse rhs not allowed"); }
/********** p = 0 ********************/   
       subs[0] = 0;  
       subs[1] = 1;
       index = mxCalcSingleSubscript(prhs[0],nsubs,subs); 
       blk_cell_pr = mxGetCell(prhs[0],index);
       numblk  = mxGetN(blk_cell_pr);
       blksize = (int)*mxGetPr(blk_cell_pr);
       if (numblk > 1) { 
          mexErrMsgTxt("mexpsqmr: only single SDP block is allowed."); }
       /*--------- get A from cell array -------------*/
       subs[0] = 0;  
       subs[1] = 0;
       index = mxCalcSingleSubscript(prhs[1],nsubs,subs); 
       A_cell_pr = mxGetCell(prhs[1],index); 
       A  = mxGetPr(A_cell_pr); 
       mA = mxGetM(A_cell_pr); 
       nA = mxGetN(A_cell_pr);
       isspA = mxIsSparse(A_cell_pr);               
       if (isspA) { irA = mxGetIr(A_cell_pr); 
                    jcA = mxGetJc(A_cell_pr); }        
       /*--------- P1,P2, P1t,P2t-------------*/
       index = mxCalcSingleSubscript(prhs[3],nsubs,subs); 
       P1_cell_pr = mxGetCell(prhs[3],index); 
       if (P1_cell_pr == NULL) { 
	  nP1 = 0; 
       } else { 
          P1 = mxGetPr(P1_cell_pr);       
          nP1 = mxGetN(P1_cell_pr); 
       }
       if (nP1 > 0) { 
          if (mxGetM(P1_cell_pr)!=blksize) { 
  	      mexErrMsgTxt("mexpsqmr: P1 and blksize is not compatible "); }  
          if (mxIsSparse(P1_cell_pr)) { 
              mexErrMsgTxt("mexpsqmr: P1 must be dense"); }
       }
       index = mxCalcSingleSubscript(prhs[4],nsubs,subs); 
       P2_cell_pr = mxGetCell(prhs[4],index); 
       if (P2_cell_pr == NULL) { 
	  nP2 = 0; 
       } else { 
          P2 = mxGetPr(P2_cell_pr);       
          nP2 = mxGetN(P2_cell_pr); 
       }
       if (nP2 > 0) { 
          if (mxGetM(P2_cell_pr)!=blksize) { 
  	      mexErrMsgTxt("mexpsqmr: P2 and blksize is not compatible "); }     
          if (mxIsSparse(P2_cell_pr)) { 
              mexErrMsgTxt("mexpsqmr: P2 must be dense"); }
       }
       index = mxCalcSingleSubscript(prhs[5],nsubs,subs); 
       Dsch12_cell_pr = mxGetCell(prhs[5],index); 
       if (Dsch12_cell_pr == NULL) { 
	  nDsch12 = 0; 
       } else { 
          Dsch12 = mxGetPr(Dsch12_cell_pr);       
          nDsch12 = mxGetN(Dsch12_cell_pr); 
       }
       if (nDsch12 > 0) {
          if (mxGetM(Dsch12_cell_pr)!=nP1 || mxGetN(Dsch12_cell_pr)!=nP2) { 
  	      mexErrMsgTxt("mexpsqmr: Dsch12 and P1,P2 is not compatible "); }     
          if (mxIsSparse(Dsch12_cell_pr)) { 
              mexErrMsgTxt("mexpsqmr: Dsch12 must be dense"); }
       }
/*************** p = 1 *************************/
    nDsch12_lin = 0; 
    if (existlinblk) {
       subs[0] = 1;  
       subs[1] = 0;
       index = mxCalcSingleSubscript(prhs[1],nsubs,subs); 
       A_cell_pr = mxGetCell(prhs[1],index); 
       A1  = mxGetPr(A_cell_pr); 
       mA1 = mxGetM(A_cell_pr); 
       nA1 = mxGetN(A_cell_pr);
       isspA1 = mxIsSparse(A_cell_pr);               
       if (isspA1) { irA1 = mxGetIr(A_cell_pr); 
                     jcA1 = mxGetJc(A_cell_pr); }       
       if (nA1 != nA) { 
          mexErrMsgTxt("mexpsqmr: blocks of A must have same number of columns."); } 
       index = mxCalcSingleSubscript(prhs[5],nsubs,subs); 
       Dsch12_cell_pr = mxGetCell(prhs[5],index); 
       if (Dsch12_cell_pr == NULL) { 
	  nDsch12_lin = 0; 
       } else {
          nDsch12_lin = mxGetM(Dsch12_cell_pr);
          Dsch12_lin = mxGetPr(Dsch12_cell_pr); 
       }
    }
/*--------- temporary arrays -------------*/
    tmpvec = mxCalloc(blksize*(blksize+1)/2,sizeof(double)); 
    matvectmp = mxCalloc(mb,sizeof(double));     
    if (existlinblk) { 
       tmpvec_lin = mxCalloc(mA1,sizeof(double)); 
    }
    if (nP1 > 0) { 
       tmparr[0] = mxCreateDoubleMatrix(nP1,blksize,mxREAL); 
       P1t = mxGetPr(tmparr[0]);
       for (j=0; j<nP1; j++) { 
         jm = j*blksize; 
         for (k=0; k<blksize; k++) { P1t[j+k*nP1] = P1[k+jm]; }     
       }
    }
    if (nP2 > 0) { 
       tmparr[1] = mxCreateDoubleMatrix(nP2,blksize,mxREAL); 
       P2t = mxGetPr(tmparr[1]); 
       for (j=0; j<nP2; j++) { 
         jm = j*blksize; 
         for (k=0; k<blksize; k++) { P2t[j+k*nP2] = P2[k+jm]; }
       }
    }
    tmparr[2] = mxCreateDoubleMatrix(blksize,blksize,mxREAL);  
    Atymat = mxGetPr(tmparr[2]); 
    if (nP1 > 0 & nP1 <= blksize/2) {
       tmparr[3] = mxCreateDoubleMatrix(nP1,blksize,mxREAL);  
       T0 = mxGetPr(tmparr[3]); 
       tmparr[4] = mxCreateDoubleMatrix(nP1,blksize,mxREAL);  
       T1  = mxGetPr(tmparr[4]); 
       tmparr[5] = mxCreateDoubleMatrix(nP1,blksize,mxREAL);  
       T2  = mxGetPr(tmparr[5]); 
       tmparr[6] = mxCreateDoubleMatrix(nP1,blksize,mxREAL);  
       Tmp = mxGetPr(tmparr[6]); 
       type = 1; 
    } else if (nP1 > blksize/2 & nP1 < blksize) {
       tmparr[3] = mxCreateDoubleMatrix(nP2,blksize,mxREAL);  
       T0 = mxGetPr(tmparr[3]); 
       tmparr[4] = mxCreateDoubleMatrix(nP2,blksize,mxREAL);  
       T1  = mxGetPr(tmparr[4]); 
       tmparr[5] = mxCreateDoubleMatrix(nP2,blksize,mxREAL);  
       T2  = mxGetPr(tmparr[5]); 
       tmparr[6] = mxCreateDoubleMatrix(nP2,blksize,mxREAL);  
       Tmp = mxGetPr(tmparr[6]); 
       type = 2; 	  
    } else if (nP1 == blksize) {
       type = 3; 
    }
/*****************************************/

   diagmatM = mxGetPr(prhs[6]); 
   if (mxGetN(prhs[6])>1) { 
      mexErrMsgTxt("mexpsqmr: diagmatM must be a column vector."); }
   pertdiag = mxGetPr(prhs[7]); 
   regparam = *mxGetPr(prhs[8]); 
   tol      = *mxGetPr(prhs[9]); 
   maxit    = (int)*mxGetPr(prhs[10]);
   stagnate_check = 20; 
   if (nrhs==12) { stagnate_check = (int)*mxGetPr(prhs[11]); }

   plhs[0] = mxCreateDoubleMatrix(mb,1,mxREAL); 
   xvec    = mxGetPr(plhs[0]); 
   plhs[1] = mxCreateDoubleMatrix(maxit,1,mxREAL); 
   resnrm  = mxGetPr(plhs[1]); 
   plhs[2] = mxCreateDoubleMatrix(1,1,mxREAL);
   solve_ok = mxGetPr(plhs[2]);
   solve_ok[0] = 1; 

   rvec = mxCalloc(mb,sizeof(double));   
   pvec = mxCalloc(mb,sizeof(double)); 
   zvec = mxCalloc(mb,sizeof(double));
   dvec = mxCalloc(mb,sizeof(double));
   res  = mxCalloc(mb,sizeof(double)); 
   Ad   = mxCalloc(mb,sizeof(double));
   Ap   = mxCalloc(mb,sizeof(double)); 
   
/******************************
*     symqmr
******************************/  

   for (j=0; j<mb; j++) { 
       rvec[j] = b[j]; 
       zvec[j] = diagmatM[j]*rvec[j]; 
       pvec[j] = zvec[j]; 
       res[j]  = rvec[j]; }    
   err = sqrt(realdot(res,res,mb)); 
   resnrm[0] = err; 

   tau_old = sqrt(realdot(zvec,zvec,mb)); 
   rho_old = realdot(rvec,pvec,mb); 
   theta_old = 0.0; 

   /*---------- main loop ------------------*/
   TINY=1e-30;    
   for (iter=1; iter<=maxit; iter++) {   
       
       for (k=0; k<mb; k++) { Ap[k] = pertdiag[k]*pvec[k]; }
       if (type==1) { 
          MATVEC1(blksize,mA,nA,A,irA,jcA,isspA, 
             nP1,nP2,Atymat,P1,P2,P1t,P2t,Dsch12,T0,T1,T2,Tmp,tmpvec,matvectmp,
             pvec,Ap,regparam); 
       } else if (type==2) {
          MATVEC2(blksize,mA,nA,A,irA,jcA,isspA, 
             nP1,nP2,Atymat,P1,P2,P1t,P2t,Dsch12,T0,T1,T2,Tmp,tmpvec,matvectmp,
             pvec,Ap,regparam); 
       } else if (type==3) {
          MATVEC3(blksize,mA,nA,A,irA,jcA,isspA, 
	     Atymat,tmpvec,matvectmp,pvec,Ap,regparam);
       }
       if (existlinblk & nDsch12_lin > 0) {
          MATVEC_lin(mA1,nA1,A1,irA1,jcA1,isspA1, 
	     Dsch12_lin,tmpvec_lin,matvectmp,pvec,Ap,regparam); 
       }
       sigma = realdot(pvec,Ap,mb); 
       if (fabs(sigma) < TINY) {
	  solve_ok[0] = 2; 
          break; 
       } else {  
	  alpha = rho_old/sigma; 
          saxpy(rvec,-alpha,Ap,mb);            
       }
       for (j=0; j<mb; j++) { zvec[j] = diagmatM[j]*rvec[j]; }

       theta = sqrt(realdot(zvec,zvec,mb))/tau_old;
       beta = 1/sqrt(1+theta*theta); 
       tau = tau_old*theta*beta; 
       gam = (beta*beta)*(theta_old*theta_old); 
       eta = (beta*beta)*alpha;       
       for (j=0; j<mb; j++) { 
           dvec[j] = gam*dvec[j]+eta*pvec[j]; }
       saxpy(xvec,1.0,dvec,mb);

       /*----- check convergence -----*/
       for (j=0; j<mb; j++) { Ad[j] = gam*Ad[j]+eta*Ap[j]; }       
       saxpy(res,-1.0,Ad,mb);
       err = sqrt(realdot(res,res,mb)); 
       resnrm[iter] = err;

       if ((err < tol) & (iter>4) & (realdot(b,xvec,mb)>0)) { break; } 
       if (iter > stagnate_check) {
   	  minratio = 1e6; 
          maxratio = 0.0;   
          for (j=0; j<9; j++) { 
              ratio = resnrm[iter-j]/resnrm[iter-j-1]; 
	      minratio = MIN(minratio,ratio); 
	      maxratio = MAX(maxratio,ratio); 
	  }
 	  if ((minratio > 0.997) & (maxratio < 1.003)) { 
   	     solve_ok[0] = -1; 
  	     printf("s"); 
	     break;  }
       }   	  
       /*-----------------------------*/
       if (fabs(rho_old) < TINY) { 
	  solve_ok[0] = 2; 
          break;  
       } else {  
	  rho = realdot(rvec,zvec,mb); 
          beta = rho/rho_old; 
          sxpay(pvec,beta,zvec,mb); 
       }
       rho_old = rho;    
       tau_old = tau; 
       theta_old = theta; 
   }  
   if (iter==maxit) { solve_ok[0] = -1; }
return;
}
/**********************************************************/



